# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2018 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import os
import io
import sys
import pytest
import time

from lib389.cli_conf.backend import (backend_list, backend_get, backend_set, backend_get_dn,
                                     backend_create, backend_delete, backend_export,
                                     backend_import, backend_add_index, backend_list_index,
                                     backend_set_index, backend_get_index, backend_reindex,
                                     backend_del_index, backend_attr_encrypt, get_monitor,
                                     backend_create_vlv, backend_del_vlv, backend_create_vlv_index,
                                     backend_get_vlv, backend_edit_vlv, backend_reindex_vlv,
                                     backend_list_vlv)

from lib389.cli_base import LogCapture, FakeArgs
from lib389.tests.cli import check_output
from lib389.tests.topologies import topology_st

from lib389.utils import ds_is_older
pytestmark = pytest.mark.skipif(ds_is_older('1.4.0'), reason="Not implemented")

stdout = sys.__stdout__
BE_NAME = 'backendRoot'
SUFFIX = 'dc=backend,dc=test'
SUB_BE_NAME = 'subBackendRoot'
SUB_SUFFIX = 'dc=sub_suffix,dc=backend,dc=test'


@pytest.fixture(scope="function")
def create_backend(topology_st, request):
    """Create backend "dc=backend,dc=test" / backendRoot
    """
    sys.stdout = io.StringIO()

    args = FakeArgs()
    args.cn = BE_NAME
    args.be_name = BE_NAME
    args.suffix = False
    args.nsslapd_suffix = SUFFIX
    args.skip_subsuffixes = False
    args.json = False
    args.parent_suffix = False
    args.create_entries = True

    args.suffix = SUFFIX
    backend_create(topology_st.standalone, None, None, args)
    check_output("The database was successfully created")

    def fin():
        sys.stdout = io.StringIO()
        args = FakeArgs()
        args.cn = BE_NAME
        args.be_name = BE_NAME
        args.suffix = SUFFIX
        args.skip_subsuffixes = False
        args.json = False

        # Delete backend
        backend_delete(topology_st.standalone, None, None, args, warn=False)
        check_output("successfully deleted")

        # Verify it's removed
        args.suffix = False
        backend_list(topology_st.standalone, None, None, args)
        check_output("backendroot", missing=True)

    request.addfinalizer(fin)


def test_backend_cli(topology_st, create_backend):
    """Test creating, listing, getting, and deleting a backend (and subsuffix)

    :id: 8a5da6ce-c24a-4d0d-9afb-dfa44ecf4145
    :setup: Standalone instance
    :steps:
        1. List backends
        2. Get backend by suffix
        3. Get backend by DN
        4. Add subsuffix
        5. Verify subsuffix
        6. Modify subsuffix
        7. Delete subsuffix
        8. Verify subsuffix is removed
        9. Modify backend
        10. Verify modify worked
        11. Test monitor works
    :expectedresults:
        1. Success
        2. Success
        3. Success
        4. Success
        5. Success
        6. Success
        7. Success
        8. Success
        9. Success
        10. Success
        11. Success
    """
    topology_st.logcap = LogCapture()
    sys.stdout = io.StringIO()

    args = FakeArgs()
    args.cn = BE_NAME
    args.be_name = BE_NAME
    args.suffix = False
    args.nsslapd_suffix = SUFFIX
    args.skip_subsuffixes = False
    args.json = False
    args.parent_suffix = False
    args.create_entries = True

    # List backend
    backend_list(topology_st.standalone, None, topology_st.logcap.log, args)
    check_output(SUFFIX)

    # Get backend by by name
    args.selector = BE_NAME
    backend_get(topology_st.standalone, None, topology_st.logcap.log, args)
    check_output(BE_NAME)

    # Get backend by DN
    args.dn = 'cn=backendRoot,cn=ldbm database,cn=plugins,cn=config'
    backend_get_dn(topology_st.standalone, None, topology_st.logcap.log, args)
    check_output(BE_NAME)

    # Add subsuffix
    args.parent_suffix = SUFFIX
    args.suffix = SUB_SUFFIX
    args.be_name = SUB_BE_NAME
    backend_create(topology_st.standalone, None, topology_st.logcap.log, args)
    check_output("The database was successfully created")

    # Verify subsuffix
    args.suffix = False
    backend_list(topology_st.standalone, None, topology_st.logcap.log, args)
    check_output(SUB_SUFFIX)

    # Modify subsuffix
    args.enable = False
    args.disable = False
    args.add_referral = False
    args.del_referral = False
    args.cache_size = False
    args.cache_memsize = False
    args.dncache_memsize = False
    args.enable_readonly = True  # Setting nsslapd-readonly to "on"
    args.disable_readonly = False
    backend_set(topology_st.standalone, None, topology_st.logcap.log, args)
    check_output("successfully updated")

    # Verify modified worked
    args.selector = SUB_BE_NAME
    backend_get(topology_st.standalone, None, topology_st.logcap.log, args)
    check_output("nsslapd-readonly: on")

    # Delete subsuffix
    args.suffix = SUB_SUFFIX
    backend_delete(topology_st.standalone, None, topology_st.logcap.log, args, warn=False)
    check_output("successfully deleted")

    # Verify it is deleted
    args.suffix = False
    backend_list(topology_st.standalone, None, topology_st.logcap.log, args)
    check_output(SUB_BE_NAME, missing=True)

    # Modify backend (use same args from subsuffix modify)
    args.be_name = BE_NAME
    backend_set(topology_st.standalone, None, topology_st.logcap.log, args)
    check_output("successfully updated")

    # Verify modified worked
    args.selector = BE_NAME
    backend_get(topology_st.standalone, None, topology_st.logcap.log, args)
    check_output("nsslapd-readonly: on")

    # Run database monitor
    args.suffix = SUFFIX
    get_monitor(topology_st.standalone, None, topology_st.logcap.log, args)
    check_output("entrycachetries")

    # Done!


def test_indexes(topology_st, create_backend):
    """Test creating, listing, getting, and deleting an index

    :id: 74e64fea-72c2-4fec-a1fe-bd2d72a075f5
    :setup: Standalone instance
    :steps:
        1. Add index (description)
        2. Verify index was added
        3. Modify index (Add type)
        4. Verify index was modified
        5. Modify index (Delete type)
        6. Verify index was modified
        7. Modify index (Add MR)
        8. Verify index was modified
        9. Modify index (Delete MR)
        10. Verify index was modified
        11. Reindex index
        12. Remove index
        13. Verify index was removed

    :expectedresults:
        1. Success
        2. Success
        3. Success
        4. Success
        5. Success
        6. Success
        7. Success
        8. Success
        9. Success
        10. Success
        11. Success
        12. Success
        13. Success
    """
    sys.stdout = io.StringIO()

    args = FakeArgs()
    args.cn = BE_NAME
    args.be_name = BE_NAME
    args.suffix = False
    args.nsslapd_suffix = SUFFIX
    args.attr = 'description'
    args.index_type = 'eq'
    args.matching_rule = None
    args.reindex = False
    args.json = False
    args.just_names = False
    args.add_type = None
    args.del_type = None
    args.add_mr = None
    args.del_mr = None

    # Add an index
    backend_add_index(topology_st.standalone, None, None, args)
    check_output("added index")

    # List indexes
    backend_list_index(topology_st.standalone, None, None, args)
    check_output("cn: description")

    # Modify index (Add type)
    args.add_type = ['sub']
    backend_set_index(topology_st.standalone, None, None, args)
    args.add_type = None
    check_output("successfully updated")

    # Verify type was added
    args.attr = ['description']
    backend_get_index(topology_st.standalone, None, None, args)
    check_output("nsindextype: sub")

    # Remove index type sub
    args.attr = 'description'

    args.del_type = ['sub']
    backend_set_index(topology_st.standalone, None, None, args)
    args.del_type = None
    check_output("successfully updated")

    # Verify type was removed
    args.attr = ['description']
    backend_get_index(topology_st.standalone, None, None, args)
    check_output("nsindextype: sub", missing=True)

    # Modify index (add MR)
    args.attr = 'description'
    args.add_mr = ['1.1.1.1.1.1']
    backend_set_index(topology_st.standalone, None, None, args)
    args.add_mr = None
    check_output("successfully updated")

    # Verify MR was added
    args.attr = ['description']
    backend_get_index(topology_st.standalone, None, None, args)
    check_output("nsmatchingrule: 1.1.1.1.1.1")

    # Modify index (delete MR)
    args.attr = 'description'
    args.del_mr = ['1.1.1.1.1.1']
    backend_set_index(topology_st.standalone, None, None, args)
    args.del_mr = None
    check_output("successfully updated")

    # Verify MR was added
    args.attr = ['description']
    backend_get_index(topology_st.standalone, None, None, args)
    check_output("nsmatchingrule: 1.1.1.1.1.1", missing=True)

    # Reindex index
    backend_reindex(topology_st.standalone, None, None, args)
    check_output("reindexed database")
    time.sleep(2)

    # Delete index
    backend_del_index(topology_st.standalone, None, None, args)
    check_output("deleted index")

    # Verify index was removed
    backend_list_index(topology_st.standalone, None, None, args)
    check_output("cn: description", missing=True)


def test_attr_encrypt(topology_st, create_backend):
    """Test adding/removing encrypted attrs

    :id: 887429d5-b9be-4e48-b8ca-691e7437abca
    :setup: Standalone instance
    :steps:
        1. Add encrypted attr
        2. Verify it succeeded
        3. Delete encrypted attr
        4. Verity it was removed
    :expectedresults:
        1. Success
        2. Success
        3. Success
        4. Success

    """
    sys.stdout = io.StringIO()
    args = FakeArgs()
    args.cn = BE_NAME
    args.be_name = BE_NAME
    args.suffix = False
    args.nsslapd_suffix = SUFFIX
    args.json = False
    args.just_names = False
    args.list = False
    args.add_attr = None
    args.del_attr = None

    # Add an encrytped attr
    args.add_attr = ['description']
    backend_attr_encrypt(topology_st.standalone, None, None, args)
    args.add_attr = None
    check_output("added encrypted attribute")

    # Verify it worked
    args.list = True
    backend_attr_encrypt(topology_st.standalone, None, None, args)
    args.list = False
    check_output("cn: description")

    # Delete encrypted attr
    args.del_attr = ['description']
    backend_attr_encrypt(topology_st.standalone, None, None, args)
    args.del_attr = None
    check_output("deleted encrypted attribute")

    # Verify it worked
    args.list = True
    backend_attr_encrypt(topology_st.standalone, None, None, args)
    args.list = False
    check_output("cn: description", missing=True)


def test_import_export(topology_st):
    BE_NAME = 'userRoot'
    EXCLUDE_SUFFIX = "ou=Groups,dc=example,dc=com"
    LDIF_NAME = "test_import_export.ldif"
    LDIF_PATH = os.path.join(topology_st.standalone.ds_paths.ldif_dir, LDIF_NAME)
    topology_st.logcap = LogCapture()
    args = FakeArgs()

    # Export the backend
    args.be_names = [BE_NAME]
    args.ldif = LDIF_NAME
    args.use_id2entry = None
    args.encrypted = None
    args.min_base64 = None
    args.no_dump_uniq_id = None
    args.replication = None
    args.not_folded = None
    args.no_seq_num = None
    args.include_suffixes = None
    args.exclude_suffixes = [EXCLUDE_SUFFIX]
    backend_export(topology_st.standalone, None, topology_st.logcap.log, args)

    # Verify export worked
    assert os.path.exists(LDIF_PATH)
    with open(LDIF_PATH, 'r') as ldif:
        for line in ldif:
            assert not line.endswith("%s\n" % EXCLUDE_SUFFIX)

    # Import the backend
    args.be_name = BE_NAME
    args.ldifs = [LDIF_NAME]
    args.chunks_size = None
    args.encrypted = None
    args.gen_uniq_id = None
    args.only_core = None
    args.include_suffixes = None
    args.exclude_suffixes = None
    backend_import(topology_st.standalone, None, topology_st.logcap.log, args)
    os.remove(LDIF_PATH)

    # Done!


def test_vlv(topology_st, create_backend):
    """Test creating, listing, getting, and deleting vlv's

    :id: b38bfcf1-f71f-4c08-9221-10f45b4c9b34
    :setup: Standalone instance
    :steps:
        1. Add VLV search and index entries
        2. Verify they are created
        3. Edit VLV search and verify change
        4. Create additional VLV indexes
        5. Verity new indedxes were created
        6. Remove VLV indexes
        7. Verify indexes were removed
        8. Reindex VLV
    :expectedresults:
        1. Success
        2. Success
        3. Success
        4. Success
        5. Success
        6. Success
        7. Success
        8. Success

    """
    sys.stdout = io.StringIO()
    args = FakeArgs()
    args.cn = BE_NAME
    args.be_name = BE_NAME
    args.suffix = False
    args.nsslapd_suffix = SUFFIX
    args.json = False
    args.name = "myVLVSearch"
    args.index_name = "myVLVIndex"
    args.search_base = SUFFIX
    args.search_scope = '2'
    args.search_filter = "cn=*"
    args.parent_name = args.name
    args.index = False
    args.reindex = False
    args.sort = "cn sn"
    args.just_names = False


    # Create vlv search
    backend_create_vlv(topology_st.standalone, None, None, args)
    check_output("created new VLV Search entry")

    # Verify search is present
    backend_get_vlv(topology_st.standalone, None, None, args)
    check_output("VLV Search:")

    # Create VLV index under vlvSearch
    backend_create_vlv_index(topology_st.standalone, None, None, args)
    check_output("created new VLV index entry")

    # Verify index is present
    backend_get_vlv(topology_st.standalone, None, None, args)
    check_output("VLV Index:")

    # Edit VLV Search
    args.search_base = None
    args.search_scope = '0'
    args.search_filter = None
    args.sort = None
    backend_edit_vlv(topology_st.standalone, None, None, args)
    check_output("updated VLV search entry")

    # Verify edit was successful
    backend_get_vlv(topology_st.standalone, None, None, args)
    check_output("vlvscope: 0")

    # List vlv searches
    backend_list_vlv(topology_st.standalone, None, None, args)
    check_output("vlvbase: " + SUFFIX)

    # Add another index
    args.index_name = "my2ndVLVIndex"
    args.sort = "uid givenname"
    backend_create_vlv_index(topology_st.standalone, None, None, args)
    check_output("created new VLV index entry")

    # Verify new index was created
    backend_get_vlv(topology_st.standalone, None, None, args)
    check_output("vlvsort: uid givenname")

    # Reindex VLV
    backend_reindex_vlv(topology_st.standalone, None, None, args)
    check_output("reindexed VLV indexes")
    time.sleep(2)

    # Delete VLV search and indexes
    backend_del_vlv(topology_st.standalone, None, None, args)
    check_output("deleted VLV search and its indexes")

    # List vlv searches/indexes
    backend_list_vlv(topology_st.standalone, None, None, args)
    check_output("")

    # Done!



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2026 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import ldap
import json
import re
from lib389._mapped_object_lint import DSLint
from lib389.cli_base import connect_instance, disconnect_instance
from lib389.cli_base.dsrc import dsrc_to_ldap, dsrc_arg_concat
from lib389.backend import Backends
from lib389.config import Encryption, Config
from lib389.monitor import MonitorDiskSpace
from lib389.replica import Replica
from lib389.nss_ssl import NssSsl
from lib389.dseldif import FSChecks, DSEldif
from lib389.dirsrv_log import DirsrvAccessLog
from lib389.tunables import Tunables
from lib389 import lint
from lib389 import plugins
from lib389._constants import DSRC_HOME
from functools import partial
from typing import Iterable


# These get single instances and check them.
CHECK_OBJECTS = [
    Config,
    Backends,
    Encryption,
    FSChecks,
    plugins.ReferentialIntegrityPlugin,
    plugins.MemberOfPlugin,
    MonitorDiskSpace,
    Replica,
    DSEldif,
    NssSsl,
    DirsrvAccessLog,
    Tunables,
]

# Checks that need a running server or LDAP search of database content (see healthcheck.md).
_OFFLINE_SKIP_SPECS_EXACT = frozenset({
    'agmts_status',
    'conflicts',
    'no_ruv',
})


def _offline_skip_spec(spec: str) -> bool:
    if spec in _OFFLINE_SKIP_SPECS_EXACT:
        return True
    if spec.endswith(':search'):
        return True
    return False


def _format_check_output(log, result, idx):
    log.info(f"\n\n[{idx}] DS Lint Error: {result['dsle']}")
    log.info("-" * 80)
    log.info(f"Severity: {result['severity']}")
    if 'check' in result:
        log.info(f"Check: {result['check']}")
    log.info("Affects:")
    for item in result['items']:
        log.info(f" -- {item}")
    log.info("\nDetails:")
    log.info("-----------")
    log.info(result['detail'])
    log.info("\nResolution:")
    log.info("-----------")
    log.info(result['fix'])


def _list_targets(inst):
    for c in CHECK_OBJECTS:
        o = c(inst)
        yield o.lint_uid(), o


def _list_errors(log):
    for r in map(partial(getattr, lint),
                 filter(partial(re.match, r'^DS'),
                        dir(lint))):
        log.info(f"{r['dsle']} :: {r['description']}")


def _list_checks(inst, specs: Iterable[str]):
    if specs is None:
        yield []
        return
    o_uids = dict(_list_targets(inst))
    for s in specs:
        wanted, rest = DSLint._dslint_parse_spec(s)
        if wanted == '*':
            raise ValueError('Unexpected spec selector asterisk')

        if wanted in o_uids:
            for l in o_uids[wanted].lint_list(rest):
                yield o_uids[wanted], l
        else:
            raise ValueError('No such object specifier: ' + wanted)


def _print_checks(inst, log, specs: Iterable[str]) -> None:
    for o, s in _list_checks(inst, specs):
        log.info(f'{o.lint_uid()}:{s[0]}')


def _run(inst, log, args, checks, exclude_checks, offline=False):
    if not args.json:
        log.info("Beginning lint report, this could take a while ...")

    report = []
    skipped = []
    excludes = []
    for _, skip in exclude_checks:
        excludes.append(skip[0])

    for o, s in checks:
        if s[0] in excludes:
            continue

        if not args.json:
            log.info(f"Checking {o.lint_uid()}:{s[0]} ...")

        if offline and _offline_skip_spec(s[0]):
            reason = 'requires running directory server or LDAP-accessible database content'
            skipped.append({'check': s[0], 'reason': reason})
            if not args.json:
                log.info(f" - skipped {s[0]} - requires running server: {reason}")
            continue

        try:
            report += o.lint(s[0]) or []
        except ValueError as e:
            if offline and 'not ONLINE' in str(e):
                reason = 'instance is not connected (offline); check needs LDAP'
                skipped.append({'check': s[0], 'reason': reason})
                if not args.json:
                    log.info(f" - skipped {s[0]} - requires running server: {reason}")
            else:
                raise
        except Exception as e:
            log.debug(f"Error running check {o.lint_uid()}:{s[0]}: {e}")
            continue

    if not args.json:
        log.info("Healthcheck complete.")

    count = len(report)
    if count == 0:
        if not args.json:
            log.info("No issues found.")
        else:
            _emit_json(log, report, skipped, [], args)
    else:
        plural = ""
        if count > 1:
            plural = "s"
        if not args.json:
            log.info(f"{count} Issue{plural} found!  Generating report ...")
            idx = 1
            for item in report:
                _format_check_output(log, item, idx)
                idx += 1
            log.info(f'\n\n===== End Of Report ({count} Issue{plural} found) =====')
        else:
            _emit_json(log, report, skipped, [], args)


def _emit_json(log, report, skipped, host_errors, args):
    """Emit JSON: list-only for backward compatibility, or wrapped object if extra sections."""
    if skipped or host_errors:
        payload = {'issues': report}
        if skipped:
            payload['skipped_requires_server'] = skipped
        if host_errors:
            payload['host_errors'] = host_errors
        log.info(json.dumps(payload, indent=4))
    else:
        log.info(json.dumps(report, indent=4))


def health_check_run(inst, log, args):
    """Connect to the local server using LDAPI, and perform various health checks.

    If the server is stopped, ``connect_instance`` fails with ``SERVER_DOWN``; we keep
    the pre-allocated ``inst`` from ``dsctl`` and run checks that do not require LDAP,
    skipping others with explicit messages. Mode is fixed for the duration of the run.
    """

    if args.list_errors:
        _list_errors(log)
        return

    # update the args for connect_instance()
    args.basedn = None
    args.binddn = None
    args.bindpw = None
    args.starttls = None
    args.pwdfile = None
    args.prompt = False
    dsrc_inst = dsrc_to_ldap(DSRC_HOME, args.instance, log.getChild('dsrc'))
    dsrc_inst = dsrc_arg_concat(args, dsrc_inst)
    offline = False
    try:
        inst = connect_instance(dsrc_inst=dsrc_inst, verbose=args.verbose,
                                args=args)
    except ldap.SERVER_DOWN:
        log.info('Directory Server instance is down — running offline-capable checks only.')
        offline = True
    except Exception as e:
        raise ValueError('Failed to connect to Directory Server instance: ' +
                         str(e)) from e

    checks = args.check or dict(_list_targets(inst)).keys()
    exclude_checks = args.exclude_check
    if args.list_checks or args.dry_run:
        _print_checks(inst, log, checks)
        return

    _run(inst, log, args, _list_checks(inst, checks),
         _list_checks(inst, exclude_checks), offline=offline)

    disconnect_instance(inst)


def create_parser(subparsers):
    run_healthcheck_parser = subparsers.add_parser('healthcheck', help=
        "Run a healthcheck report on a local Directory Server instance. This "
        "is a safe and read-only operation.  Do not attempt to run this on a "
        "remote Directory Server as this tool needs access to local resources, "
        "otherwise the report may be inaccurate.")
    run_healthcheck_parser.set_defaults(func=health_check_run)
    run_healthcheck_parser.add_argument('--list-checks', action='store_true', help='List of known checks')
    run_healthcheck_parser.add_argument('--list-errors', action='store_true', help='List of known error codes')
    run_healthcheck_parser.add_argument('--dry-run', action='store_true', help='Do not execute the actual check, only list what would be done')
    run_healthcheck_parser.add_argument('--check', nargs='+', default=None,
                                        help='Areas to check. These can be obtained by --list-checks. Every element on the left of the colon (:)'
                                             ' may be replaced by an asterisk if multiple options on the right are available.')
    run_healthcheck_parser.add_argument('--exclude-check', nargs='+', default=[],
                                        help='Areas to skip. These can be obtained by --list-checks. Every element on the left of the colon (:)'
                                             ' may be replaced by an asterisk if multiple options on the right are available.')



from lib389.utils import DSVersion
import pytest

versions = [('1.3.10.1', '1.3.2.1'),
          ('2.3.2', '1.4.4.4'),
          ('2.3.2.202302121950git1b4f5a5bf', '2.3.2'),
          ('1.1.0a', '1.1.0')]

@pytest.mark.parametrize("x,y", versions)
def test_dsversion(x, y):
    assert DSVersion(x) > DSVersion(y)




# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2016 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import os
import sys
import time
import ldap
import logging
import pytest
from lib389 import DirSrv, Entry, tools, tasks
from lib389.tools import DirSrvTools
from lib389._constants import *
from lib389.properties import *
from lib389.tasks import *
from lib389.utils import *

from lib389.plugins import *

DEBUGGING = False

if DEBUGGING:
    logging.getLogger(__name__).setLevel(logging.DEBUG)
else:
    logging.getLogger(__name__).setLevel(logging.INFO)


log = logging.getLogger(__name__)


class TopologyStandalone(object):
    """The DS Topology Class"""
    def __init__(self, standalone):
        """Init"""
        standalone.open()
        self.standalone = standalone


@pytest.fixture(scope="module")
def topology(request):
    """Create DS Deployment"""

    # Creating standalone instance ...
    if DEBUGGING:
        standalone = DirSrv(verbose=True)
    else:
        standalone = DirSrv(verbose=False)
    args_instance[SER_HOST] = HOST_STANDALONE
    args_instance[SER_PORT] = PORT_STANDALONE
    args_instance[SER_SERVERID_PROP] = SERVERID_STANDALONE
    args_instance[SER_CREATION_SUFFIX] = DEFAULT_SUFFIX
    args_standalone = args_instance.copy()
    standalone.allocate(args_standalone)
    instance_standalone = standalone.exists()
    if instance_standalone:
        standalone.delete()
    standalone.create()
    standalone.open()


    def fin():
        """If we are debugging just stop the instances, otherwise remove
        them
        """
        if DEBUGGING:
            standalone.stop(60)
        else:
            standalone.delete()

    request.addfinalizer(fin)

    # Clear out the tmp dir
    standalone.clearTmpDir(__file__)

    return TopologyStandalone(standalone)


def test_plugin_management(topology):
    """
    This tests the new plugin management and enablement features.
    """


    if DEBUGGING:
        # Add debugging steps(if any)...
        pass

    for plugin in topology.standalone.plugins.list():
        print(plugin)


    # Make sure that when we get the plugin for refint, it's the correct
    # type
    refintplugin = topology.standalone.plugins.get('libreferint-plugin')
    assert(isinstance(refintplugin, ReferentialIntegrityPlugin))

    # Set and configure a simple plugin (IE no extra config)

    rolesplugin = topology.standalone.plugins.get('libroles-plugin')
    assert(isinstance(rolesplugin, Plugin))

    # Neither of these should fail
    rolesplugin.disable()
    rolesplugin.enable()

    # Prove that we can set and configure a complex plugin

    uniqplugin = topology.standalone.plugins.get('attribute uniqueness')
    assert(isinstance(uniqplugin, AttributeUniquenessPlugin))

    uniqplugin.disable()
    uniqplugin.add_unique_attribute('sn')
    uniqplugin.enable_all_subtrees()
    uniqplugin.enable()

    # Test compat with the old api

    topology.standalone.plugins.enable('attribute uniqueness')

    # Show the status

    assert(uniqplugin.status())

    log.info('Test PASSED')


if __name__ == '__main__':
    # Run isolated
    # -s for DEBUG mode
    CURRENT_FILE = os.path.realpath(__file__)
    pytest.main("-s %s" % CURRENT_FILE)





# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2015 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---
#
import os
from lib389 import Entry
import lib389
import pytest

import logging
logging.basicConfig(level=logging.DEBUG)
log = logging.getLogger(__name__)


class TestEntry(object):
    """A properly initialized Entry:
        - accepts well-formed or empty dn and tuples;
        - refuses empty dn
    """
    def test_init_empty(self):
        e = Entry('')
        assert not e.dn

    def test_init_with_str(self):
        e = Entry('o=pippo')
        assert e.dn == 'o=pippo'

    def test_init_badstr(self):
        # This should not be allowed
        with pytest.raises(ValueError):
            e = Entry('no equal sign here')

    def test_init_with_tuple(self):
        expected = 'pippo'
        given = 'o=pippo'
        t = (given, {
             'o': [expected],
             'objectclass': ['organization', 'top']
             })
        e = Entry(t)
        assert e.dn == given
        assert expected in e.o

    def test_update(self):
        expected = 'pluto minnie'
        given = {'cn': expected}
        t = ('o=pippo', {
             'o': ['pippo'],
             'objectclass': ['organization', 'top']
             })

        e = Entry(t)
        e.update(given)
        assert e.cn == expected, "Bad cn: %s, expected: %s" % (e.cn, expected)

    def test_update_complex(self):
        # compare two entries created with different methods
        nsuffix, replid, replicatype = ("dc=example,dc=com", 5,
                                        lib389.REPLICA_RDWR_TYPE)
        binddnlist, legacy = ['uid=pippo, cn=config'], 'off'
        dn = "dc=example,dc=com"
        entry = Entry(dn)
        entry.setValues(
            'objectclass', "top", "nsds5replica", "extensibleobject")
        entry.setValues('cn', "replica")
        entry.setValues('nsds5replicaroot', nsuffix)
        entry.setValues('nsds5replicaid', str(replid))
        entry.setValues('nsds5replicatype', str(replicatype))
        entry.setValues('nsds5flags', "1")
        entry.setValues('nsds5replicabinddn', binddnlist)
        entry.setValues('nsds5replicalegacyconsumer', legacy)

        uentry = Entry((
            dn, {'objectclass': ["top", "nsds5replica", "extensibleobject"],
                 'cn': ["replica"]})
        )
        log.debug("Entry created with dict:", uentry)
        # Entry.update *replaces*, so be careful with multi-valued attrs
        uentry.update({
            'nsds5replicaroot': nsuffix,
            'nsds5replicaid': str(replid),
            'nsds5replicatype': str(replicatype),
            'nsds5flags': '1',
            'nsds5replicabinddn': binddnlist,
            'nsds5replicalegacyconsumer': legacy
        })
        uentry_s, entry_s = list(map(str, (uentry, entry)))
        assert uentry_s == entry_s, "Mismatching entries [%r] vs [%r]" % (
            uentry, entry)


if __name__ == "__main__":
    CURRENT_FILE = os.path.realpath(__file__)
    pytest.main("-s -v %s" % CURRENT_FILE)



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2017 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---
#
import ldap

import logging
import sys
import time

import pytest

from lib389 import DirSrv
from lib389._constants import *
from lib389.properties import *

from lib389.tests.topologies import topology_st

from lib389.utils import ds_is_older
pytestmark = pytest.mark.skipif(ds_is_older('1.4.0'), reason="Not implemented")

REQUIRED_DNS = [
    'dc=example,dc=com',
    'ou=Groups,dc=example,dc=com',
    'ou=People,dc=example,dc=com',
    'ou=Special Users,dc=example,dc=com',
    'cn=Accounting Managers,ou=Groups,dc=example,dc=com',
    'cn=HR Managers,ou=Groups,dc=example,dc=com',
    'cn=QA Managers,ou=Groups,dc=example,dc=com',
    'cn=PD Managers,ou=Groups,dc=example,dc=com',
    'cn=Directory Administrators,dc=example,dc=com',
]

def test_install_sample_entries(topology_st):
    # Assert that our entries match.

    entries = topology_st.standalone.search_s(DEFAULT_SUFFIX, ldap.SCOPE_SUBTREE)

    for entry in entries:
        assert(entry.dn in REQUIRED_DNS)
        # We can make this assert the full object content, plugins and more later.

    assert(True)



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2021 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---
#
import ldap
import time
import os
import pytest

from lib389 import NoSuchEntryError
from lib389.agreement import Agreement
from lib389._constants import *
from lib389.properties import *
from lib389 import DirSrv
from lib389.utils import ensure_bytes, ensure_str
from lib389.idm.user import UserAccounts, TEST_USER_PROPERTIES
from lib389.idm.domain import Domain

# Used for One supplier / One consumer topology
HOST_SUPPLIER = LOCALHOST
PORT_SUPPLIER = 40389
SERVERID_SUPPLIER = 'supplier'
REPLICAID_SUPPLIER = 1

HOST_CONSUMER = LOCALHOST
PORT_CONSUMER = 50389
SERVERID_CONSUMER = 'consumer'

SUFFIX = DEFAULT_SUFFIX
ENTRY_DN = "cn=test_entry, %s" % SUFFIX
SECOND_AGMT_TEST_PORT = 12345


class TopologyReplication(object):
    def __init__(self, supplier, consumer):
        supplier.open()
        consumer.open()
        self.supplier = supplier
        self.consumer = consumer


@pytest.fixture(scope="module")
def topology(request):
    # Supplier
    #
    # Create the supplier instance
    supplier = DirSrv(verbose=False)
    supplier.log.debug("supplier allocated")
    args = {SER_HOST: HOST_SUPPLIER,
            SER_PORT: PORT_SUPPLIER,
            SER_SERVERID_PROP: SERVERID_SUPPLIER}
    supplier.allocate(args)
    if supplier.exists():
        supplier.delete()
    supplier.create()
    supplier.open()

    # Enable replication
    supplier.replica.enableReplication(suffix=SUFFIX,
                                       role=ReplicaRole.SUPPLIER,
                                       replicaId=REPLICAID_SUPPLIER)

    # Consumer
    #
    # Create the consumer instance
    consumer = DirSrv(verbose=False)
    consumer.log.debug("Consumer allocated")
    args = {SER_HOST: HOST_CONSUMER,
            SER_PORT: PORT_CONSUMER,
            SER_SERVERID_PROP: SERVERID_CONSUMER}
    consumer.allocate(args)
    if consumer.exists():
        consumer.delete()
    consumer.create()
    consumer.open()

    # Enable replication
    consumer.replica.enableReplication(suffix=SUFFIX,
                                       role=ReplicaRole.CONSUMER)

    # Delete each instance in the end
    def fin():
        supplier.delete()
        consumer.delete()
    request.addfinalizer(fin)

    return TopologyReplication(supplier, consumer)


def test_create(topology):
    """Test to create a replica agreement and initialize the consumer.
    Test on a unknown suffix
    """

    topology.supplier.log.info("\n\n##############\n## CREATE\n##############\n")
    properties = {RA_NAME: ('meTo_%s:%d' % (topology.consumer.host,
                                            topology.consumer.port)),
                  RA_BINDDN: defaultProperties[REPLICATION_BIND_DN],
                  RA_BINDPW: defaultProperties[REPLICATION_BIND_PW],
                  RA_METHOD: defaultProperties[REPLICATION_BIND_METHOD],
                  RA_TRANSPORT_PROT: defaultProperties[REPLICATION_TRANSPORT]}
    repl_agreement = topology.supplier.agreement.create(
        suffix=SUFFIX, host=topology.consumer.host,
        port=topology.consumer.port, properties=properties)
    topology.supplier.log.debug("%s created" % repl_agreement)
    topology.supplier.agreement.init(SUFFIX, HOST_CONSUMER, PORT_CONSUMER)
    topology.supplier.waitForReplInit(repl_agreement)

    supplier_users = UserAccounts(topology.supplier, SUFFIX)
    consumer_users = UserAccounts(topology.consumer, SUFFIX)

    testuser = supplier_users.create(properties=TEST_USER_PROPERTIES)
    testuser_dn = testuser.dn

    # Add a test entry
    # Check replication is working
    loop = 0
    while loop <= 10:
        try:
            consumer_users.get(dn=testuser_dn)
            break
        except ldap.NO_SUCH_OBJECT:
            time.sleep(1)
            loop += 1
    assert loop <= 10

    # Check that with an invalid suffix it raises NoSuchEntryError
    with pytest.raises(NoSuchEntryError):
        properties = {RA_NAME: r'meAgainTo_%s:%d' %
                      (topology.consumer.host, topology.consumer.port)}
        topology.supplier.agreement.create(suffix="ou=dummy",
                                           host=topology.consumer.host,
                                           port=topology.consumer.port,
                                           properties=properties)


def test_list(topology):
    """List the replica agreement on a suffix => 1
    Add a RA
    List the replica agreements on that suffix again => 2
    List a specific RA

    PREREQUISITE: it exists a replica for SUFFIX and a replica agreement
    """

    topology.supplier.log.info("\n\n###########\n## LIST\n#############\n")
    ents = topology.supplier.agreement.list(suffix=SUFFIX)
    assert len(ents) == 1
    assert ents[0].getValue(RA_PROPNAME_TO_ATTRNAME[RA_CONSUMER_HOST]) == \
        ensure_bytes(topology.consumer.host)
    assert ents[0].getValue(RA_PROPNAME_TO_ATTRNAME[RA_CONSUMER_PORT]) == \
        ensure_bytes(str(topology.consumer.port))

    # Create a second RA to check .list returns 2 RA
    properties = {RA_NAME: r'meTo_%s:%d' % (topology.consumer.host,
                                            SECOND_AGMT_TEST_PORT),
                  RA_BINDDN: defaultProperties[REPLICATION_BIND_DN],
                  RA_BINDPW: defaultProperties[REPLICATION_BIND_PW],
                  RA_METHOD: defaultProperties[REPLICATION_BIND_METHOD],
                  RA_TRANSPORT_PROT: defaultProperties[REPLICATION_TRANSPORT]}
    topology.supplier.agreement.create(suffix=SUFFIX,
                                       host=topology.consumer.host,
                                       port=SECOND_AGMT_TEST_PORT,
                                       properties=properties)
    ents = topology.supplier.agreement.list(suffix=SUFFIX)
    assert len(ents) == 2

    # Check we can .list a specific RA
    ents = topology.supplier.agreement.list(suffix=SUFFIX,
                                            consumer_host=topology.consumer.host,
                                            consumer_port=topology.consumer.port)
    assert len(ents) == 1
    assert ents[0].getValue(RA_PROPNAME_TO_ATTRNAME[RA_CONSUMER_HOST]) == \
        ensure_bytes(topology.consumer.host)
    assert ents[0].getValue(RA_PROPNAME_TO_ATTRNAME[RA_CONSUMER_PORT]) == \
        ensure_bytes(str(topology.consumer.port))


def test_delete(topology):
    """Delete previously created the replica agreement

    PREREQUISITE: it exists a replica for SUFFIX and a replica agreement
    with a SECOND_AGMT_TEST_PORT specified
    """

    # TODO: add a few test cases for different sets of arguments
    #       using fixtures and parameters.
    topology.supplier.log.info("\n\n##############\n### DELETE\n##############\n")
    ents = topology.supplier.agreement.list(suffix=SUFFIX,
                                            consumer_host=topology.consumer.host,
                                            consumer_port=SECOND_AGMT_TEST_PORT)
    assert len(ents) == 1

    # Find DN of the second agreement
    replica_entries = topology.supplier.replica.list(SUFFIX)
    replica = replica_entries[0]
    agreement_cn = r'meTo_%s:%d' % (topology.consumer.host,
                                    SECOND_AGMT_TEST_PORT)
    agreement_dn = ','.join(["cn=%s" % agreement_cn, replica.dn])

    topology.supplier.agreement.delete(suffix=SUFFIX,
                                       consumer_host=topology.consumer.host,
                                       consumer_port=SECOND_AGMT_TEST_PORT,
                                       agmtdn=agreement_dn)

    ents = topology.supplier.agreement.list(suffix=SUFFIX,
                                            consumer_host=topology.consumer.host,
                                            consumer_port=SECOND_AGMT_TEST_PORT)
    assert not ents


def test_status(topology):
    """Test that status is returned from agreement"""

    topology.supplier.log.info("\n\n###########\n## STATUS\n##########")
    ents = topology.supplier.agreement.list(suffix=SUFFIX)
    for ent in ents:
        ra_status = topology.supplier.agreement.status(ensure_str(ent.dn))
        assert ra_status
        topology.supplier.log.info("Status of %s: %s" % (ent.dn, ra_status))


def test_schedule(topology):
    """Test the schedule behaviour with valid and invalid values"""

    topology.supplier.log.info("\n\n###########\n## SCHEDULE\n#########")
    ents = topology.supplier.agreement.list(suffix=SUFFIX,
                                            consumer_host=topology.consumer.host,
                                            consumer_port=topology.consumer.port)
    assert len(ents) == 1

    topology.supplier.agreement.schedule(ents[0].dn, Agreement.ALWAYS)
    ents = topology.supplier.agreement.list(suffix=SUFFIX,
                                            consumer_host=topology.consumer.host,
                                            consumer_port=topology.consumer.port)
    assert len(ents) == 1
    assert ents[0].getValue(RA_PROPNAME_TO_ATTRNAME[RA_SCHEDULE]) == \
        ensure_bytes(Agreement.ALWAYS)

    topology.supplier.agreement.schedule(ents[0].dn, Agreement.NEVER)
    ents = topology.supplier.agreement.list(suffix=SUFFIX,
                                            consumer_host=topology.consumer.host,
                                            consumer_port=topology.consumer.port)
    assert len(ents) == 1
    assert ents[0].getValue(RA_PROPNAME_TO_ATTRNAME[RA_SCHEDULE]) == \
        ensure_bytes(Agreement.NEVER)

    CUSTOM_SCHEDULE = "0000-1234 6420"
    topology.supplier.agreement.schedule(ents[0].dn, CUSTOM_SCHEDULE)
    ents = topology.supplier.agreement.list(suffix=SUFFIX,
                                            consumer_host=topology.consumer.host,
                                            consumer_port=topology.consumer.port)
    assert len(ents) == 1
    assert ents[0].getValue(RA_PROPNAME_TO_ATTRNAME[RA_SCHEDULE]) == \
        ensure_bytes(CUSTOM_SCHEDULE)

    CUSTOM_SCHEDULES = ("2500-1234 6420",  # Invalid HOUR schedule
                        "0000-2534 6420",  # ^^
                        "1300-1234 6420",  # Starting HOUR after ending HOUR
                        "0062-1234 6420",  # Invalid MIN schedule
                        "0000-1362 6420",  # ^^
                        "0000-1234 6-420",  # Invalid DAYS schedule
                        "0000-1362 64209",  # ^^
                        "0000-1362 01234560")  # ^^

    for CUSTOM_SCHEDULE in CUSTOM_SCHEDULES:
        with pytest.raises(ValueError):
            topology.supplier.agreement.schedule(ents[0].dn, CUSTOM_SCHEDULE)


def test_getProperties(topology):
    """Check the correct behaviour of getProperties function"""

    topology.supplier.log.info("\n\n###########\n## GETPROPERTIES\n############")
    ents = topology.supplier.agreement.list(suffix=SUFFIX,
                                            consumer_host=topology.consumer.host,
                                            consumer_port=topology.consumer.port)
    assert len(ents) == 1
    properties = topology.supplier.agreement.getProperties(agmnt_dn=ents[0].dn)
    for prop in properties:
        topology.supplier.log.info("RA %s : %s -> %s" %
                                   (prop, RA_PROPNAME_TO_ATTRNAME[prop],
                                    properties[prop]))

    properties = topology.supplier.agreement.getProperties(
        agmnt_dn=ents[0].dn, properties=[RA_BINDDN])
    assert len(properties) == 1
    for prop in properties:
        topology.supplier.log.info("RA %s : %s -> %s" %
                                   (prop, RA_PROPNAME_TO_ATTRNAME[prop],
                                    properties[prop]))


def test_setProperties(topology):
    """Set properties to the agreement and check, if it was successful"""

    topology.supplier.log.info("\n\n###########\n### SETPROPERTIES\n##########")
    ents = topology.supplier.agreement.list(suffix=SUFFIX,
                                            consumer_host=topology.consumer.host,
                                            consumer_port=topology.consumer.port)
    assert len(ents) == 1
    test_schedule = "1234-2345 12345"
    test_desc = "test_desc"
    topology.supplier.agreement.setProperties(
        agmnt_dn=ents[0].dn, properties={RA_SCHEDULE: test_schedule,
                                         RA_DESCRIPTION: test_desc})
    properties = topology.supplier.agreement.getProperties(
        agmnt_dn=ents[0].dn, properties=[RA_SCHEDULE, RA_DESCRIPTION])
    assert len(properties) == 2
    assert properties[RA_SCHEDULE][0] == ensure_bytes(test_schedule)
    assert properties[RA_DESCRIPTION][0] == ensure_bytes(test_desc)

    # Set RA Schedule back to "always"
    topology.supplier.agreement.schedule(ents[0].dn, Agreement.ALWAYS)


def test_changes(topology):
    """Test the changes counter behaviour after making some changes
    to the replicated suffix
    """

    topology.supplier.log.info("\n\n##########\n### CHANGES\n##########")
    ents = topology.supplier.agreement.list(suffix=SUFFIX,
                                            consumer_host=topology.consumer.host,
                                            consumer_port=topology.consumer.port)
    assert len(ents) == 1
    value = topology.supplier.agreement.changes(agmnt_dn=ents[0].dn)
    topology.supplier.log.info("\ntest_changes: %d changes\n" % value)
    assert value > 0

    # Do an update
    TEST_STRING = 'test_string'
    supplier_domain = Domain(topology.supplier, SUFFIX)
    consumer_domain = Domain(topology.consumer, SUFFIX)

    supplier_domain.set('description', TEST_STRING)

    # The update has been replicated
    loop = 0
    while loop <= 10:
        if consumer_domain.present('description', TEST_STRING):
            break
        time.sleep(1)
        loop += 1
    assert loop <= 10

    # Give a little time to update a change number on supplier
    time.sleep(2)

    # Check change number
    newvalue = topology.supplier.agreement.changes(agmnt_dn=ents[0].dn)
    topology.supplier.log.info("\ntest_changes: %d changes\n" % newvalue)
    assert (value + 1) == newvalue


if __name__ == "__main__":
    CURRENT_FILE = os.path.realpath(__file__)
    pytest.main("-s -v %s" % CURRENT_FILE)



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2017 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---
#

import pytest
import ldap

from lib389.tests.topologies import topology_st
from lib389.plugins import ReferentialIntegrityPlugin

from lib389.lint import *

def test_hc_backend_mt(topology_st):
    mt = topology_st.standalone.mappingtrees.get('userRoot')
    # We have to remove the MT from a backend.
    mt.delete()
    be = topology_st.standalone.backends.get('userRoot')
    result = be._lint_mappingtree()
    # We have to check this by name, not object, as the result changes
    # the affected dn in the result.
    assert result['dsle'] == 'DSBLE0001'

def test_hc_encryption(topology_st):
    topology_st.standalone.encryption.set('sslVersionMin', 'TLS1.0')
    result = topology_st.standalone.encryption._lint_check_tls_version()
    assert result == DSELE0001

def test_hc_config(topology_st):
    # Check the password scheme check.
    topology_st.standalone.config.set('passwordStorageScheme', 'SSHA')
    result = topology_st.standalone.config._lint_passwordscheme()
    assert result == DSCLE0002

def test_hc_referint(topology_st):
    plugin = ReferentialIntegrityPlugin(topology_st.standalone)
    plugin.enable()

    # Assert we don't get an error when delay is 0.
    plugin.set('referint-update-delay', '0')
    result = plugin._lint_update_delay()
    assert result is None

    # Assert we get an error when delay is not 0.
    plugin.set('referint-update-delay', '10')
    result = plugin._lint_update_delay()
    assert result == DSRILE0001

    # Assert we don't get an error when plugin is disabled.
    plugin.disable()
    result = plugin._lint_update_delay()
    assert result is None



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2017 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---
#

# Implement types for COS handling with lib389 and DS.


from lib389._mapped_object import DSLdapObject, DSLdapObjects

from lib389.utils import ensure_str


class CosTemplate(DSLdapObject):
    """A Cos Template defining the values to override on a target.

    :param instance: DirSrv instance
    :type instance: DirSrv
    :param dn: The dn of the template
    :type dn: str
    """

    def __init__(self, instance, dn=None):
        super(CosTemplate, self).__init__(instance, dn)
        self._rdn_attribute = 'cn'
        self._must_attributes = ['cn']
        # This is the ONLY TIME i'll allow extensible object ...
        # You have been warned ...
        self._create_objectclasses = [
            'top',
            'cosTemplate',
            'extensibleObject',
        ]
        self._protected = False


class CosTemplates(DSLdapObjects):
    """The set of costemplates that exist for direct and indirect
    implementations.

    :param instance: A dirsrv instance
    :type instance: DirSrv
    :param basedn: The basedn of the templates
    :type basedn: str
    :param rdn: The rdn of the templates
    :type rdn: str
    """

    def __init__(self, instance, basedn, rdn=None):
        super(CosTemplates, self).__init__(instance)
        self._objectclasses = [
            'cosTemplate'
        ]
        self._filterattrs = ['cn']
        self._childobject = CosTemplate
        self._basedn = basedn
        if rdn is not None:
            self._basedn = '{},{}'.format(ensure_str(rdn), ensure_str(basedn))


class CosIndirectDefinition(DSLdapObject):
    """A Cos Indirect Definition associating an attr:value pair as a link
    attr to a template type.

    :param instance: DirSrv instance
    :type instance: DirSrv
    :param dn: The dn of the template
    :type dn: str
    """

    def __init__(self, instance, dn=None):
        super(CosIndirectDefinition, self).__init__(instance, dn)
        self._rdn_attribute = 'cn'
        self._must_attributes = ['cn', 'cosIndirectSpecifier', 'cosattribute']
        self._create_objectclasses = [
            'top',
            'ldapsubentry',
            'cosSuperDefinition',
            'cosIndirectDefinition',
        ]
        self._protected = False


class CosIndirectDefinitions(DSLdapObjects):
    """The set of cos indirect definitions that exist.

    :param instance: A dirsrv instance
    :type instance: DirSrv
    :param basedn: The basedn of the templates
    :type basedn: str
    :param rdn: The rdn of the templates
    :type rdn: str
    """

    def __init__(self, instance, basedn, rdn=None):
        super(CosIndirectDefinitions, self).__init__(instance)
        self._objectclasses = [
            'top',
            'ldapsubentry',
            'cosSuperDefinition',
            'cosIndirectDefinition',
        ]
        self._filterattrs = ['cn']
        self._childobject = CosIndirectDefinition
        self._basedn = basedn
        if rdn is not None:
            self._basedn = '{},{}'.format(ensure_str(rdn), ensure_str(basedn))


class CosPointerDefinition(DSLdapObject):
    """A Cos Pointer Definition associating a dn syntax type as a link
    attr to a template type.

    :param instance: DirSrv instance
    :type instance: DirSrv
    :param dn: The dn of the template
    :type dn: str
    """

    def __init__(self, instance, dn=None):
        super(CosPointerDefinition, self).__init__(instance, dn)
        self._rdn_attribute = 'cn'
        self._must_attributes = ['cn', 'cosTemplateDn', 'cosAttribute']
        self._create_objectclasses = [
            'top',
            'ldapsubentry',
            'cosSuperDefinition',
            'cosPointerDefinition',
        ]
        self._protected = False


class CosPointerDefinitions(DSLdapObjects):
    """The set of cos pointer definitions that exist.

    :param instance: A dirsrv instance
    :type instance: DirSrv
    :param basedn: The basedn of the templates
    :type basedn: str
    :param rdn: The rdn of the templates
    :type rdn: str
    """

    def __init__(self, instance, basedn, rdn=None):
        super(CosPointerDefinitions, self).__init__(instance)
        self._objectclasses = [
            'top',
            'ldapsubentry',
            'cosSuperDefinition',
            'cosPointerDefinition',
        ]
        self._filterattrs = ['cn']
        self._childobject = CosPointerDefinition
        self._basedn = basedn
        if rdn is not None:
            self._basedn = '{},{}'.format(ensure_str(rdn), ensure_str(basedn))


class CosClassicDefinition(DSLdapObject):
    """A Cos CosClassicDefinition associating a dn syntax type as a link
        attr to a template type.

        :param instance: DirSrv instance
        :type instance: DirSrv
        :param dn: The dn of the template
        :type dn: str
        """
    def __init__(self, instance, dn=None):
        super(CosClassicDefinition, self).__init__(instance, dn)
        self._rdn_attribute = 'cn'
        self._must_attributes = ['cn']
        self._create_objectclasses = [
            'top',
            'ldapsubentry',
            'cossuperdefinition',
            'cosClassicDefinition',
        ]
        self._protected = False


class CosClassicDefinitions(DSLdapObjects):
    """The set of cos CosClassicDefinition  that exist.
    :param instance: A dirsrv instance
    :type instance: DirSrv
    :param basedn: The basedn of the templates
    :type basedn: str
    :param rdn: The rdn of the templates
    :type rdn: str
    """

    def __init__(self, instance, basedn):
        super(CosClassicDefinitions, self).__init__(instance)
        self._objectclasses = [
            'top',
            'ldapsubentry',
            'cossuperdefinition',
            'cosClassicDefinition',
        ]
        self._filterattrs = ['cn']
        self._childobject = CosClassicDefinition
        self._basedn = basedn



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2021 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---


import io
import sys
import pytest
import time
import json
from lib389.cli_base import LogCapture, FakeArgs
from lib389.utils import *
from lib389._constants import *
from lib389.idm.nscontainer import nsContainers
from lib389.tests.topologies import topology_m2 as topo
from lib389.cli_conf.conflicts import (list_conflicts, cmp_conflict, del_conflict, swap_conflict,
                                       convert_conflict, list_glue, del_glue, convert_glue)
from lib389.utils import ds_is_older
pytestmark = pytest.mark.skipif(ds_is_older('1.4.0'), reason="Not implemented")


def _create_container(inst, dn, name):
    """Creates container entry"""
    containers = nsContainers(inst, dn)
    container = containers.create(properties={'cn': name})
    time.sleep(1)
    return container


def _delete_container(container):
    """Deletes container entry"""
    container.delete()
    time.sleep(1)


def test_conflict_cli(topo):
    """Test manageing replication conflict entries

    :id: 800f432a-52ab-4661-ac66-a2bdd9b984d8
    :setup: two supplers
    :steps:
        1. Create replication conflict entries
        2. List conflicts
        3. Compare conflict entry
        4. Delete conflict
        5. Resurrect conflict
        6. Swap conflict
        7. List glue entry
        8. Delete glue entry
        9. Convert glue entry

    :expectedresults:
        1. Success
        2. Success
        3. Success
        4. Success
        5. Success
        6. Success
        7. Success
        8. Success
        9. Success
        10. Success
    """

    # Setup our default parameters for CLI functions
    topo.logcap = LogCapture()
    sys.stdout = io.StringIO()
    args = FakeArgs()
    args.DN = ""
    args.suffix = DEFAULT_SUFFIX
    args.json = True

    m1 = topo.ms["supplier1"]
    m2 = topo.ms["supplier2"]

    topo.pause_all_replicas()

    # Create entries
    _create_container(m1, DEFAULT_SUFFIX, 'conflict_parent1')
    _create_container(m2, DEFAULT_SUFFIX, 'conflict_parent1')
    _create_container(m1, DEFAULT_SUFFIX, 'conflict_parent2')
    _create_container(m2, DEFAULT_SUFFIX, 'conflict_parent2')
    cont_parent_m1 = _create_container(m1, DEFAULT_SUFFIX, 'conflict_parent3')
    cont_parent_m2 = _create_container(m2, DEFAULT_SUFFIX, 'conflict_parent3')
    cont_glue_m1 = _create_container(m1, DEFAULT_SUFFIX, 'conflict_parent4')
    cont_glue_m2 = _create_container(m2, DEFAULT_SUFFIX, 'conflict_parent4')

    # Create the conflicts
    _delete_container(cont_parent_m1)
    _create_container(m2, cont_parent_m2.dn, 'conflict_child1')
    _delete_container(cont_glue_m1)
    _create_container(m2, cont_glue_m2.dn, 'conflict_child2')

    # Resume replication
    topo.resume_all_replicas()
    time.sleep(5)

    # Test "list"
    list_conflicts(m2, None, topo.logcap.log, args)
    conflicts = json.loads(topo.logcap.outputs[0].getMessage())
    assert len(conflicts['items']) == 4
    conflict_1_DN = conflicts['items'][0]['dn']
    conflict_2_DN = conflicts['items'][1]['dn']
    conflict_3_DN = conflicts['items'][2]['dn']
    topo.logcap.flush()

    # Test compare
    args.DN = conflict_1_DN
    cmp_conflict(m2, None, topo.logcap.log, args)
    conflicts = json.loads(topo.logcap.outputs[0].getMessage())
    assert len(conflicts['items']) == 2
    topo.logcap.flush()

    # Test delete
    del_conflict(m2, None, topo.logcap.log, args)
    list_conflicts(m2, None, topo.logcap.log, args)
    conflicts = json.loads(topo.logcap.outputs[0].getMessage())
    assert len(conflicts['items']) == 3
    topo.logcap.flush()

    # Test swap
    args.DN = conflict_2_DN
    swap_conflict(m2, None, topo.logcap.log, args)
    list_conflicts(m2, None, topo.logcap.log, args)
    conflicts = json.loads(topo.logcap.outputs[0].getMessage())
    assert len(conflicts['items']) == 2
    topo.logcap.flush()

    # Test conflict convert
    args.DN = conflict_3_DN
    args.new_rdn = "cn=testing convert"
    convert_conflict(m2, None, topo.logcap.log, args)
    list_conflicts(m2, None, topo.logcap.log, args)
    conflicts = json.loads(topo.logcap.outputs[0].getMessage())
    assert len(conflicts['items']) == 1
    topo.logcap.flush()

    # Test list glue entries
    list_glue(m2, None, topo.logcap.log, args)
    glues = json.loads(topo.logcap.outputs[0].getMessage())
    assert len(glues['items']) == 2
    topo.logcap.flush()

    # Test delete glue entries
    args.DN = "cn=conflict_parent3,dc=example,dc=com"
    del_glue(m2, None, topo.logcap.log, args)
    list_glue(m2, None, topo.logcap.log, args)
    glues = json.loads(topo.logcap.outputs[0].getMessage())
    assert len(glues['items']) == 1
    topo.logcap.flush()

    # Test convert glue entries
    args.DN = "cn=conflict_parent4,dc=example,dc=com"
    convert_glue(m2, None, topo.logcap.log, args)
    list_glue(m2, None, topo.logcap.log, args)
    glues = json.loads(topo.logcap.outputs[0].getMessage())
    assert len(glues['items']) == 0
    topo.logcap.flush()



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2015 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---
#
from lib389._constants import *
from lib389 import DirSrv, Entry
import ldap
import pytest
import logging

logging.getLogger(__name__).setLevel(logging.DEBUG)
log = logging.getLogger(__name__)

INSTANCE_PORT = 54321
INSTANCE_SERVERID = 'standalone'


class TopologyStandalone(object):
    def __init__(self, standalone):
        standalone.open()
        self.standalone = standalone


@pytest.fixture(scope='module')
def topology(request):
    standalone = DirSrv(verbose=False)
    standalone.log.debug('Instance allocated')
    args = {SER_HOST: LOCALHOST,
            SER_PORT: INSTANCE_PORT,
            SER_SERVERID_PROP: INSTANCE_SERVERID}
    standalone.allocate(args)
    if standalone.exists():
        standalone.delete()
    standalone.create()
    standalone.open()

    def fin():
        standalone.delete()
    request.addfinalizer(fin)

    return TopologyStandalone(standalone)


@pytest.fixture(scope='module')
def user(topology):
    """Create user entries"""

    for i in range(0, 2):
        uentry = Entry('uid=test%s,%s' % (i, DEFAULT_SUFFIX))
        uentry.setValues('objectclass', 'top', 'extensibleobject')
        uentry.setValues('uid', 'test')
        topology.standalone.add_s(uentry)


@pytest.fixture(scope='module')
def group(topology):
    """Create a group entry"""

    gentry = Entry('cn=testgroup,%s' % DEFAULT_SUFFIX)
    gentry.setValues('objectclass', 'top', 'extensibleobject')
    gentry.setValues('cn', 'testgroup')
    for i in range(0, 2):
        gentry.setValues('uniqueMember', 'uid=test%s,%s' % (i, DEFAULT_SUFFIX))
    topology.standalone.add_s(gentry)


def test_dereference(topology, user, group):
    """Test dereference search argument formating
    Check, that result is correct
    """

    log.info('Check, that our deref Control Value is correctly formatted')
    with pytest.raises(ldap.UNAVAILABLE_CRITICAL_EXTENSION):
        result, control_response = topology.standalone.dereference(
            'uniqueMember:dn,uid;uniqueMember:dn,uid',
            filterstr='(cn=testgroup)')

    result, control_response = topology.standalone.dereference(
        'uniqueMember:cn,uid,objectclass', filterstr='(cn=testgroup)')

    log.info('Check, that the dereference search result is correct')
    assert result[0][2][0].entry == [{'derefVal':
                                      'uid=test1,dc=example,dc=com',
                                      'derefAttr': 'uniqueMember',
                                      'attrVals': [{'vals':
                                                    ['top',
                                                     'extensibleobject'],
                                                    'type': 'objectclass'},
                                                   {'vals': ['test', 'test1'],
                                                    'type': 'uid'}]}]


if __name__ == '__main__':
    CURRENT_FILE = os.path.realpath(__file__)
    pytest.main('-s -v %s' % CURRENT_FILE)



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2021 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---
#
import os
import ldap
import pytest
import logging

from lib389 import NoSuchEntryError
from lib389.replica import Replicas
from lib389.backend import Backends
from lib389.idm.domain import Domain
from lib389._constants import (ReplicaRole, BACKEND_SUFFIX, BACKEND_NAME, REPLICA_RUV_FILTER, CONSUMER_REPLICAID,
                               REPLICA_FLAGS_WRITE, REPLICA_RDWR_TYPE)
from lib389.properties import (REPL_FLAGS, REPL_TYPE)
from lib389.tests.topologies import topology_i3 as topo

logging.getLogger(__name__).setLevel(logging.DEBUG)
log = logging.getLogger(__name__)

DEBUGGING = os.getenv('DEBUGGING', default=False)
NEW_SUFFIX = 'dc=test,dc=com'
NEW_BACKEND = 'test_backend'
REPLICA_SUPPLIER_ID = 1


@pytest.fixture(scope="module")
def new_suffixes(topo):
    """Create new suffix, backend and mapping tree"""

    for num in range(1, 4):
        backends = Backends(topo.ins["standalone{}".format(num)])
        backends.create(properties={BACKEND_SUFFIX: NEW_SUFFIX,
                                    BACKEND_NAME: NEW_BACKEND})
        domain = Domain(topo.ins["standalone{}".format(num)], NEW_SUFFIX)
        domain.create(properties={'dc': 'test', 'description': NEW_SUFFIX})


@pytest.fixture(scope="function")
def simple_replica(topo, new_suffixes, request):
    """Enable simple multi-supplier replication"""

    supplier1 = topo.ins["standalone1"]
    supplier2 = topo.ins["standalone2"]

    log.info("Enable two supplier replicas")
    replicas_m1 = Replicas(supplier1)
    replica_m1 = replicas_m1.enable(suffix=NEW_SUFFIX,
                                    role=ReplicaRole.SUPPLIER,
                                    replicaID=REPLICA_SUPPLIER_ID)
    replicas_m2 = Replicas(supplier2)
    replica_m2 = replicas_m2.enable(suffix=NEW_SUFFIX,
                                    role=ReplicaRole.SUPPLIER,
                                    replicaID=REPLICA_SUPPLIER_ID+1)

    log.info("Create agreements between the instances")
    supplier1.agreement.create(suffix=NEW_SUFFIX,
                               host=supplier2.host,
                               port=supplier2.port)
    supplier2.agreement.create(suffix=NEW_SUFFIX,
                               host=supplier1.host,
                               port=supplier1.port)

    log.info("Test replication")
    replicas_m1.test(NEW_SUFFIX, supplier2)

    def fin():
            replicas_m1.disable(NEW_SUFFIX)
            replicas_m2.disable(NEW_SUFFIX)
    request.addfinalizer(fin)

    return [replica_m1, replica_m2]


@pytest.fixture(scope="function")
def clean_up(topo, new_suffixes, request):
    """Check that all replicas were disabled and disable if not"""

    def fin():
        for num in range(1, 4):
            try:
                replicas = Replicas(topo.ins["standalone{}".format(num)])
                replicas.disable(NEW_SUFFIX)
                log.info("standalone{} is disabled now".format(num))
            except:
                pass
    request.addfinalizer(fin)


def test_delete_agreements(topo, simple_replica):
    """Check deleteAgreements method

    :feature: Replication
    :steps: 1. Enable replication with agreements
            2. Delete the agreements
            3. Check that agreements were deleted
            4. Disable replication
    :expectedresults: No errors happen, agreements successfully deleted
    """

    supplier1 = topo.ins["standalone1"]
    supplier2 = topo.ins["standalone2"]

    log.info("Check that agreements in place")
    ents = supplier1.agreement.list(suffix=NEW_SUFFIX)
    assert(len(ents) == 1)
    ents = supplier2.agreement.list(suffix=NEW_SUFFIX)
    assert(len(ents) == 1)

    log.info("Delete the agreements")
    simple_replica[0].deleteAgreements()
    simple_replica[1].deleteAgreements()

    log.info("Check that agreements were deleted")
    ents = supplier1.agreement.list(suffix=NEW_SUFFIX)
    assert(len(ents) == 0)
    ents = supplier2.agreement.list(suffix=NEW_SUFFIX)
    assert(len(ents) == 0)


def test_get_ruv_entry(topo, simple_replica):
    """Check get_ruv_entry method

    :feature: Replication
    :steps: 1. Enable replication with agreements
            2. Get ruv entry with get_ruv_entry() method
            3. Get ruv entry with ldap.search
            4. Disable replication
    :expectedresults: Entries should be equal
    """

    ruv_entry = simple_replica[0].get_ruv_entry()
    entry = topo.ins["standalone1"].search_s(NEW_SUFFIX, ldap.SCOPE_SUBTREE, REPLICA_RUV_FILTER)[0]

    assert ruv_entry == entry


def test_get_role(topo, simple_replica):
    """Check get_role method

    :feature: Replication
    :steps: 1. Enable replication with agreements
            2. Get role with get_role() method
            3. Get repl_flags, repl_type from the replica entry with ldap.search
            4. Compare the values
            5. Disable replication
    :expectedresults: The role 'supplier' should have flags=1 and type=3
    """

    replica_role = simple_replica[0].get_role()
    replica_flags = simple_replica[0].get_attr_val_int(REPL_FLAGS)
    replica_type = simple_replica[0].get_attr_val_int(REPL_TYPE)

    log.info("Check that we've got role 'supplier', while {}=1 and {}=3".format(REPL_FLAGS, REPL_TYPE))
    assert replica_role == ReplicaRole.supplier and replica_flags == REPLICA_FLAGS_WRITE \
        and replica_type == REPLICA_RDWR_TYPE, \
        "Failure, get_role() gave {}, while {} has {} and {} has {}".format(replica_role,
                                                                            REPL_FLAGS, replica_flags,
                                                                            REPL_TYPE, replica_type)


def test_basic(topo, new_suffixes, clean_up):
    """Check basic replica functionality

    :feature: Replication
    :steps: 1. Enable replication on supplier. hub and consumer
            2. Create agreements: supplier-hub, hub-consumer
            3. Test supplier-consumer replication
            4. Disable replication
            5. Check that replica, agreements and changelog were deleted
    :expectedresults: No errors happen, replication is successfully enabled and disabled
    """

    supplier = topo.ins["standalone1"]
    hub = topo.ins["standalone2"]
    consumer = topo.ins["standalone3"]

    log.info("Enable replicas (create replica and changelog entries)")
    supplier_replicas = Replicas(supplier)
    supplier_replicas.enable(suffix=NEW_SUFFIX,
                             role=ReplicaRole.SUPPLIER,
                             replicaID=REPLICA_SUPPLIER_ID)
    ents = supplier_replicas.list()
    assert len(ents) == 1
    ents = supplier.changelog.list()
    assert len(ents) == 1

    hub_replicas = Replicas(hub)
    hub_replicas.enable(suffix=NEW_SUFFIX,
                        role=ReplicaRole.HUB,
                        replicaID=CONSUMER_REPLICAID)
    ents = hub_replicas.list()
    assert len(ents) == 1
    ents = hub.changelog.list()
    assert len(ents) == 1

    consumer_replicas = Replicas(consumer)
    consumer_replicas.enable(suffix=NEW_SUFFIX,
                             role=ReplicaRole.CONSUMER)
    ents = consumer_replicas.list()
    assert len(ents) == 1

    log.info("Create agreements between the instances")
    supplier.agreement.create(suffix=NEW_SUFFIX,
                              host=hub.host,
                              port=hub.port)
    ents = supplier.agreement.list(suffix=NEW_SUFFIX)
    assert len(ents) == 1
    hub.agreement.create(suffix=NEW_SUFFIX,
                         host=consumer.host,
                         port=consumer.port)
    ents = hub.agreement.list(suffix=NEW_SUFFIX)
    assert len(ents) == 1

    log.info("Test replication")
    supplier_replicas.test(NEW_SUFFIX, consumer)

    log.info("Disable replication")
    supplier_replicas.disable(suffix=NEW_SUFFIX)
    hub_replicas.disable(suffix=NEW_SUFFIX)
    consumer_replicas.disable(suffix=NEW_SUFFIX)

    log.info("Check that replica, agreements and changelog were deleted")
    for num in range(1, 4):
        log.info("Checking standalone{} instance".format(num))
        inst = topo.ins["standalone{}".format(num)]

        log.info("Checking that replica entries don't exist")
        replicas = Replicas(inst)
        ents = replicas.list()
        assert len(ents) == 0

        log.info("Checking that changelog doesn't exist")
        ents = inst.changelog.list()
        assert len(ents) == 0

        log.info("Checking that agreements can't be acquired because the replica entry doesn't exist")
        with pytest.raises(NoSuchEntryError) as e:
            inst.agreement.list(suffix=NEW_SUFFIX)
            assert "no replica set up" in e.msg


@pytest.mark.parametrize('role_from,role_to',
                         ((ReplicaRole.CONSUMER, ReplicaRole.HUB),
                          (ReplicaRole.CONSUMER, ReplicaRole.SUPPLIER),
                          (ReplicaRole.HUB, ReplicaRole.SUPPLIER)))
def test_promote(topo, new_suffixes, clean_up, role_from, role_to):
    """Check that replica promote method works properly

    :feature: Replication
    :steps: 1. Enable replication on the instance
            2. Promote it to another role
               (check consumer-hub, consumer-supplier, hub-supplier
            3. Check that role was successfully changed
            4. Disable replication
    :expectedresults: No errors happen, replica successfully promoted
    """

    inst = topo.ins["standalone1"]

    log.info("Enable replication on instance with a role - {}".format(role_from))
    replicas = Replicas(inst)
    replica = replicas.enable(suffix=NEW_SUFFIX,
                              role=role_from)

    log.info("Promote replica to {}".format(role_to))
    replica.promote(newrole=role_to,
                    rid=REPLICA_SUPPLIER_ID)

    log.info("Check that replica was successfully promoted")
    replica_role = replica.get_role()
    assert replica_role == role_to


@pytest.mark.parametrize('role_from,role_to',
                         ((ReplicaRole.SUPPLIER, ReplicaRole.HUB),
                          (ReplicaRole.SUPPLIER, ReplicaRole.CONSUMER),
                          (ReplicaRole.HUB, ReplicaRole.CONSUMER)))
def test_demote(topo, new_suffixes, clean_up, role_from, role_to):
    """Check that replica demote method works properly

    :feature: Replication
    :steps: 1. Enable replication on the instance
            2. Demote it to another role
               (check supplier-hub, supplier-consumer, hub-consumer)
            3. Check that role was successfully changed
            4. Disable replication
    :expectedresults: No errors happen, replica successfully demoted
    """

    inst = topo.ins["standalone1"]

    log.info("Enable replication on instance with a role - {}".format(role_from.name))
    replicas = Replicas(inst)
    replica = replicas.enable(suffix=NEW_SUFFIX,
                              role=role_from,
                              replicaID=REPLICA_SUPPLIER_ID)

    log.info("Promote replica to {}".format(role_to.name))
    replica.demote(newrole=role_to)

    log.info("Check that replica was successfully promoted")
    replica_role = replica.get_role()
    assert replica_role == role_to


@pytest.mark.parametrize('role_from', (ReplicaRole.SUPPLIER,
                                       ReplicaRole.HUB,
                                       ReplicaRole.CONSUMER))
def test_promote_fail(topo, new_suffixes, clean_up, role_from):
    """Check that replica promote method fails
    when promoted to wrong direction

    :feature: Replication
    :steps: 1. Enable replication on the instance
            2. Try to promote it to wrong role
               (for example, supplier-hub, hub-consumer)
            3. Disable replication
    :expectedresults: Replica shouldn't be promoted
    """

    inst = topo.ins["standalone1"]

    log.info("Enable replication on instance with a role - {}".format(role_from.name))
    replicas = Replicas(inst)
    replica = replicas.enable(suffix=NEW_SUFFIX,
                              role=role_from,
                              replicaID=REPLICA_SUPPLIER_ID)

    for role_to in [x for x in range(1, 4) if x <= role_from.value]:
        role_to = ReplicaRole(role_to)
        log.info("Try to promote replica to {}".format(role_to.name))
        with pytest.raises(ValueError):
            replica.promote(newrole=role_to,
                            rid=REPLICA_SUPPLIER_ID)


@pytest.mark.parametrize('role_from', (ReplicaRole.SUPPLIER,
                                       ReplicaRole.HUB,
                                       ReplicaRole.CONSUMER))
def test_demote_fail(topo, new_suffixes, clean_up, role_from):
    """Check that replica demote method fails
    when demoted to wrong direction

    :feature: Replication
    :steps: 1. Enable replication on the instance
            2. Try to demote it to wrong role
               (for example, consumer-supplier, hub-supplier)
            3. Disable replication
    :expectedresults: Replica shouldn't be demoted
    """

    inst = topo.ins["standalone1"]

    log.info("Enable replication on instance with a role - {}".format(role_from.name))
    replicas = Replicas(inst)
    replica = replicas.enable(suffix=NEW_SUFFIX,
                              role=role_from,
                              replicaID=REPLICA_SUPPLIER_ID)

    for role_to in [x for x in range(1, 4) if x >= role_from.value]:
        role_to = ReplicaRole(role_to)
        log.info("Try to demote replica to {}".format(role_to.name))
        with pytest.raises(ValueError):
            replica.demote(newrole=role_to)


if __name__ == "__main__":
    CURRENT_FILE = os.path.realpath(__file__)
    pytest.main("-s -v %s" % CURRENT_FILE)



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2019 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import ldap
from lib389.plugins import ReferentialIntegrityPlugin, ReferentialIntegrityConfig
from lib389.cli_conf import add_generic_plugin_parsers, generic_object_edit, generic_object_add
from lib389.cli_base import CustomHelpFormatter

arg_to_attr = {
    'update_delay': 'referint-update-delay',
    'membership_attr': 'referint-membership-attr',
    'entry_scope': 'nsslapd-pluginEntryScope',
    'exclude_entry_scope': 'nsslapd-pluginExcludeEntryScope',
    'container_scope': 'nsslapd-pluginContainerScope',
    'config_entry': 'nsslapd-pluginConfigArea',
    'log_file': 'referint-logfile'
}


def referint_edit(inst, basedn, log, args):
    log = log.getChild('referint_edit')
    plugin = ReferentialIntegrityPlugin(inst)
    generic_object_edit(plugin, log, args, arg_to_attr)


def referint_add_config(inst, basedn, log, args):
    log = log.getChild('referint_add_config')
    targetdn = args.DN
    config = generic_object_add(ReferentialIntegrityConfig, inst, log, args, arg_to_attr, dn=targetdn)
    plugin = ReferentialIntegrityPlugin(inst)
    plugin.replace('nsslapd-pluginConfigArea', config.dn)
    log.info('ReferentialIntegrity attribute nsslapd-pluginConfigArea (config-entry) '
             'was set in the main plugin config')


def referint_edit_config(inst, basedn, log, args):
    log = log.getChild('referint_edit_config')
    targetdn = args.DN
    config = ReferentialIntegrityConfig(inst, targetdn)
    generic_object_edit(config, log, args, arg_to_attr)


def referint_show_config(inst, basedn, log, args):
    log = log.getChild('referint_show_config')
    targetdn = args.DN
    config = ReferentialIntegrityConfig(inst, targetdn)

    if not config.exists():
        raise ldap.NO_SUCH_OBJECT("Entry %s doesn't exists" % targetdn)
    if args and args.json:
        o_str = config.get_all_attrs_json()
        log.info(o_str)
    else:
        log.info(config.display())


def referint_del_config(inst, basedn, log, args):
    log = log.getChild('referint_del_config')
    targetdn = args.DN
    config = ReferentialIntegrityConfig(inst, targetdn)
    config.delete()
    # Now remove the attribute from the plugin
    plugin = ReferentialIntegrityPlugin(inst)
    plugin.remove_all('nsslapd-pluginConfigArea')
    log.info("Successfully deleted the %s", targetdn)


def _add_parser_args(parser):
    parser.add_argument('--update-delay',
                        help='Sets the update interval. Special values: 0 - The check is performed immediately, '
                             '-1 - No check is performed (referint-update-delay)')
    parser.add_argument('--membership-attr', nargs='+',
                        help='Specifies attributes to check for and update (referint-membership-attr)')
    parser.add_argument('--entry-scope',
                        help='Defines the subtree in which the plug-in looks for the delete '
                             'or rename operations of a user entry (nsslapd-pluginEntryScope)')
    parser.add_argument('--exclude-entry-scope',
                        help='Defines the subtree in which the plug-in ignores any operations '
                             'for deleting or renaming a user (nsslapd-pluginExcludeEntryScope)')
    parser.add_argument('--container-scope',
                        help='Specifies which branch the plug-in searches for the groups to which the user belongs. '
                             'It only updates groups that are under the specified container branch, '
                             'and leaves all other groups not updated (nsslapd-pluginContainerScope)')
    parser.add_argument('--log-file',
                        help='Specifies a path to the Referential integrity logfile.'
                             'For example: /var/log/dirsrv/slapd-YOUR_INSTANCE/referint')


def create_parser(subparsers):
    referint = subparsers.add_parser('referential-integrity',
                                      help='Manage and configure Referential Integrity Postoperation plugin')

    subcommands = referint.add_subparsers(help='action')

    add_generic_plugin_parsers(subcommands, ReferentialIntegrityPlugin)

    edit = subcommands.add_parser('set', help='Edit the plugin settings', formatter_class=CustomHelpFormatter)
    edit.set_defaults(func=referint_edit)
    _add_parser_args(edit)
    edit.add_argument('--config-entry', help='The value to set as nsslapd-pluginConfigArea')

    config = subcommands.add_parser('config-entry', help='Manage the config entry', formatter_class=CustomHelpFormatter)
    config_subcommands = config.add_subparsers(help='action')
    add_config = config_subcommands.add_parser('add', help='Add the config entry', formatter_class=CustomHelpFormatter)
    add_config.set_defaults(func=referint_add_config)
    add_config.add_argument('DN', help='The config entry full DN')
    _add_parser_args(add_config)
    edit_config = config_subcommands.add_parser('set', help='Edit the config entry', formatter_class=CustomHelpFormatter)
    edit_config.set_defaults(func=referint_edit_config)
    edit_config.add_argument('DN', help='The config entry full DN')
    _add_parser_args(edit_config)
    show_config = config_subcommands.add_parser('show', help='Display the config entry', formatter_class=CustomHelpFormatter)
    show_config.set_defaults(func=referint_show_config)
    show_config.add_argument('DN', help='The config entry full DN')
    del_config_ = config_subcommands.add_parser('delete',
                                                help='Delete the config entry and remove the reference in the plugin entry',
                                                formatter_class=CustomHelpFormatter)
    del_config_.set_defaults(func=referint_del_config)
    del_config_.add_argument('DN', help='The config entry full DN')
